#current_path <- getwd()
#parent_path <- dirname(current_path)
#setwd(parent_path)

#load("ADNI_data.RData")
library("MASS")
library("fda")
library("glmnet")
library("Matrix")


X_minus_AH <- function(X, A, H) { 
  X_minus_AH <- array(0, c(n, p, T) )
  for (t in 1 : T) X_minus_AH[, , t] <- X[, , t] - H[, , t] %*% t(A)
  return(X_minus_AH)
}

AB_array_multi <- function(A, B) {
  AB <- matrix(0, length(A[1, , 1]), length(B[1, , 1]) )
  for (t in 1 : T) AB <- AB + t(A[, , t]) %*% B[, , t]
  return(AB)
}

B_ite <- function(X, f, H, HH_inv, lambda) {
  X_minus_fH <- X_minus_AH(X, f, H)
  XH_fH <- AB_array_multi(X_minus_fH, H) 
  B <- XH_fH %*% HH_inv
  B <- apply(B, 2, function(x) x <- sign(x) * ifelse( (abs(x) - lambda) > 0, (abs(x) - lambda), 0) )
  return(B)
}

Alpha_ite <- function(X, B, H, HH_inv, Bs, BBs_inv) {
  X_minus_BH <- X_minus_AH(X, B, H)
  XH_BH <- AB_array_multi(X_minus_BH, H) 
  Alpha <- HH_inv %*% t(XH_BH) %*% Bs %*% BBs_inv
  return(Alpha)
}

Theta_score_ite <- function(Wl, K) {
  Wl_svd <- svd( t(Wl) %*% Wl )
  Theta_q_l <- t( Wl_svd$u[, 1 : K] ) 
  score_n_l <- Wl %*% t(Theta_q_l)
  pro_l <- sum( (Wl_svd$d)[1 : K] ) / sum( (Wl_svd$d) )
  return(list(Theta_q_l, score_n_l, pro_l) )
}

tau_Bs_k <- 4
break_num_s <- 3
tau_Bs <- tau_Bs_k^3 
Bs1 <- bsplineS(x = xyz[, 1],seq(0, 1, length = break_num_s), tau_Bs_k + 2 - break_num_s)
Bs2 <- bsplineS(x = xyz[, 2],seq(0, 1, length = break_num_s), tau_Bs_k + 2 - break_num_s)
Bs3 <- bsplineS(x = xyz[, 3],seq(0, 1, length = break_num_s), tau_Bs_k + 2 - break_num_s)
Bs <- matrix(t( KhatriRao( KhatriRao(t(Bs1),t(Bs2)), t(Bs3) ) ) , p, tau_Bs )
Bs <- (svd( Bs %*% t(Bs) ))$u[, 1 : tau_Bs] * sqrt(p)
for (k in 1 : tau_Bs) Bs[, k] <- sign(Bs[1, k]) * Bs[, k]
BBs_inv <- solve( t(Bs) %*% Bs ) # the last term in Alpha

tau_Bt <- 5
break_num_t <- 5
Bt <- bsplineS(time,breaks = seq(0, 1, length = break_num_t), norder = tau_Bt + 2 - break_num_t)
Bt <- (svd( Bt %*% t(Bt) ))$u[, 1 : tau_Bt] * sqrt(T)
for (k in 1 : tau_Bt) Bt[, k] <- sign(Bt[1, k]) * Bt[, k]
BBt_inv <- solve( t(Bt) %*% Bt ) # the first term in W


n <- 803
X_SFFPCA <- array(0,c(n,p,T))

X_mean <- matrix(0, p, T)
for (t in 1 : T) X_mean[, t] <- colMeans(X_all[, ,t])
for (i in 1 : n) X_SFFPCA[i, ,] <- X_all[i, , ] - X_mean 

q <- 3
K <- 2


XX <- AB_array_multi(X_SFFPCA, X_SFFPCA)
XX_svd <- svd(XX)
Bf0 <- XX_svd$u[, 1 : q] * sqrt(p)     
Alpha0 <- t( BBs_inv %*% t(Bs) %*% Bf0 )
f0 <- Bs %*% t(Alpha0)

B0 <- Bf0 - f0
H0 <- array(0, c(n, q, T))
for (i in 1 : n) H0[i, , ] <- t(Bf0) %*% X_SFFPCA[i, , ] / p 

B <- B0
f <- f0
H <- H0
lambda1 <- 0.7
lambda2 <- 12
Weight_s <- matrix(0, p, p)
for (j in 1 : p) Weight_s[j, -j] <- apply(xyz[-j, ], 1, function(x) return(1/norm(x - xyz[j, ], type="2")) )
Weight_s <- t(Weight_s)
D <- kronecker(diag(p) * (p) - matrix(rep(1, p*p) ,p ,p), diag(q) )
E <- rep(0, p*q)
mu <- 0.01
B_minus_B <-  array(0, c(p, p, q))
for (j in 1 : p ) B_minus_B[, j, ]  <- t(apply(B, 1, function(x) x - B[j, ] ))  
Gam <- B_minus_B
nu <- array(0, c(p, p, q)) 
W <- array(0, c(n, tau_Bt, q) )
phi <- array(0,c(q, K, T) )
Theta_q <- array(0, c(K, tau_Bt, q) )
score_n <- array(0, c(n, q, K) )
Pro <- matrix(0, q, step_max)

for (k in 1 : step_max) {
  
  wtH <- matrix(0, p*q, p*q)
  for (j in 1 : p) wtH[ ((j-1)*q+1) : (j*q) , ((j-1)*q+1) : (j*q) ] <- AB_array_multi(H, H) 
  X_minus_fH <- X_minus_AH(X_SFFPCA, f, H)
  XH_fH <- AB_array_multi(X_minus_fH, H) 
  wtH_X_minus_fH  <- c(t(XH_fH))
  HH_inv <- solve( AB_array_multi(H, H) )
  
  for (l in 1 : 3) {
    for (j in 1 : p) E[ ((j-1)*q+1)  :  (j*q) ] <- colSums(nu[j, -j,] - nu[-j, j, ] - Gam[j, -j, ] + Gam[-j, j, ]) 
    wtB <- solve(2 * wtH + mu * D/10) %*% (2 * wtH_X_minus_fH - E/10 )
    B <- t(matrix(wtB, q, p))
    for (j in 1 : p) {
      if ( norm(B[j, ],type = "2") - lambda1 > 0) { B[j, ] <- ( norm(B[j, ],type = "2") - lambda1 ) * B[j, ] / norm( B[j, ],type = "2") }
      else {B[j, ] <- rep(0,q)}
    }
    
    for (j in 1 : p ) B_minus_B[, j, ] <- t(apply(B, 1, function(x) x - B[j, ] ))  
    Gam <- nu / mu + B_minus_B
    for (j in 1 : p) {
      for (j1 in 1 : p) { 
        if (norm(c(Gam[j, j1, ]),type = "2") - lambda2*Weight_s[j, j1] > 0) { Gam[j, j1, ] <- (norm(Gam[j, j1, ],type = "2")-lambda2*Weight_s[j, j1]) * Gam[j, j1, ] / norm(Gam[j, j1, ],type = "2") }
        else {Gam[j, j1, ] <- rep(0,q)}
      }
    }
    nu <- nu + mu * (B_minus_B - Gam)/10
  }
  
  Alpha <- Alpha_ite(X_SFFPCA, B, H, HH_inv, Bs, BBs_inv)     
  f <- Bs %*% t(Alpha)
  
  Bf <- B+f
  for (t in 1 : T) H[, , t] = X_SFFPCA[, , t] %*% (Bf ) %*% solve( t(Bf) %*% (Bf) )
  for (i in 1 : n) W[i, , ] =  BBt_inv %*% t(Bt) %*% t(H[i, , ])
  for (l in 1 : q) {
    Theta_score_result_l <- Theta_score_ite(W[, , l], K)
    Theta_q[, , l] <- Theta_score_result_l[[1]]     
    score_n[, l, ] <- Theta_score_result_l[[2]]     
    phi[l, , ] <- t( Bt %*% t( Theta_q[, , l] ))
    H[, l, ] <-  score_n[, l, ] %*% phi[l, , ]
  }
  
}
score <- matrix(0, n, q*K)
for (i in 1 : n) score[i, ] <- c( t(score_n[i, ,]) )







data <- cbind(MMSE,gender1, age1,score)
PE_lm <- rep(0, 100)

for (seed in 1:100) {
  set.seed(seed)
  index <- sample(1:n,0.5*n,replace = FALSE)
  data_train <- data[index, ]
  data_test <- data[-index, ]
  
  lm_result <- lm(MMSE~., data = as.data.frame(data_train))
  lm_coef <- lm_result$coefficients
  lm_predict <- apply(data_test[,-1],1, function(x) x%*% lm_coef[-1] + lm_coef[1])
  pe_lm <- mean((sort(c((lm_predict - data_test[,1])^2)/var(data_test[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])
  PE_lm[seed] <- pe_lm
}




print( c("PE_lm",mean(PE_lm),sd(PE_lm))  )



dat_volume <- cbind(MMSE,gender1, age1, volume)
vol_svd <- svd(volume) 
vol_fac_2 <- vol_svd$u[,1:2]
vol_fac_10 <- vol_svd$u[,1:10]
vol_fac_20 <- vol_svd$u[,1:20]
dat_volume_fac_2 <- cbind(MMSE,gender1, age1, vol_fac_2)
dat_volume_fac_10 <- cbind(MMSE,gender1, age1, vol_fac_10)
dat_volume_fac_20 <- cbind(MMSE,gender1, age1, vol_fac_20)

PE_vol_ridge_2 <- PE_fac_lm_2 <- PE_fac_lm_10 <- PE_fac_lm_20 <- PE_fac_ridge_20 <- rep(0,100)

for (seed in 1:100) {
  set.seed(seed)
  index <- sample(1:n,0.5*n,replace = FALSE)
  data_train <- data[index, ]
  data_test <- data[-index, ]
  dat_volume_train <- dat_volume[index, ]
  dat_volume_test <- dat_volume[-index, ]
  
  # scalar volume 
  vol_ridge_result_2 <- glmnet(dat_volume_train[, -1], dat_volume_train[, 1],alpha = 0, intercept = TRUE, lambda = 5)
  vol_ridge_coef_2 <- c(vol_ridge_result_2$a0, (vol_ridge_result_2$beta)[,1])
  vol_ridge_predict_2 <- apply(dat_volume_test[,-1],1, function(x) x%*% vol_ridge_coef_2[-1] + vol_ridge_coef_2[1])
  pe_vol_ridge_2 <- mean((sort(c((vol_ridge_predict_2 - data_test[,1])^2)/var(data_test[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])

  PE_vol_ridge_2[seed] <- pe_vol_ridge_2
  
  ########### scalar factor
  data_fac_train_2 <- dat_volume_fac_2[index, ]
  data_fac_train_10 <- dat_volume_fac_10[index, ]
  data_fac_train_20 <- dat_volume_fac_20[index, ]
  data_fac_test_2 <- dat_volume_fac_2[-index, ]
  data_fac_test_10 <- dat_volume_fac_10[-index, ]
  data_fac_test_20 <- dat_volume_fac_20[-index, ] 
  
  lm_fac_result_2 <- lm(MMSE~.-1, data = as.data.frame(data_fac_train_2))
  lm_fac_coef_2 <- lm_fac_result_2$coefficients
  lm_fac_predict_2 <- apply(data_fac_test_2[,-1],1, function(x) x%*% lm_fac_coef_2)
  pe_fac_lm_2 <- mean((sort(c((lm_fac_predict_2 - data_fac_test_2[,1])^2)/var(data_fac_test_2[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])
  
  lm_fac_result_10 <- lm(MMSE~.-1, data = as.data.frame(data_fac_train_10))
  lm_fac_coef_10 <- lm_fac_result_10$coefficients
  lm_fac_predict_10 <- apply(data_fac_test_10[,-1],1, function(x) x%*% lm_fac_coef_10)
  pe_fac_lm_10 <- mean((sort(c((lm_fac_predict_10 - data_fac_test_10[,1])^2)/var(data_fac_test_10[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])
  
  
  lm_fac_result_20 <- lm(MMSE~.-1, data = as.data.frame(data_fac_train_20))
  ridge_fac_result_20 <- glmnet(data_fac_train_20[, -1], data_fac_train_20[, 1],alpha = 0, intercept = TRUE, lambda = 0.01)
  lm_fac_coef_20 <- lm_fac_result_20$coefficients
  ridge_fac_coef_20 <- c(ridge_fac_result_20$a0,(ridge_fac_result_20$beta)[,1])
  lm_fac_predict_20 <- apply(data_fac_test_20[,-1],1, function(x) x%*% lm_fac_coef_20 )
  ridge_fac_predict_20 <- apply(data_fac_test_20[,-1],1, function(x) x%*% ridge_fac_coef_20[-1] + ridge_fac_coef_20[1])
  pe_fac_lm_20 <- mean((sort(c((lm_fac_predict_20 - data_fac_test_20[,1])^2)/var(data_fac_test_20[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])
  pe_fac_ridge_20 <- mean((sort(c((ridge_fac_predict_20 - data_fac_test_20[,1])^2)/var(data_fac_test_20[,1]),decreasing = FALSE))[1:(0.95*0.5*n)])
  
  
  PE_fac_lm_2[seed] <- pe_fac_lm_2
  PE_fac_lm_10[seed] <- pe_fac_lm_10
  PE_fac_lm_20[seed] <- pe_fac_lm_20
  PE_fac_ridge_20[seed] <- pe_fac_ridge_20
}

PE_Fun_LR <- cbind(PE_lm,"Fun-LR")
PE_Vol_RR <- cbind(PE_vol_ridge_2,"Vol-RR")
PE_Vol_F2_LR <- cbind(PE_fac_lm_2,"Vol-F2-LR")
PE_Vol_F10_LR <- cbind(PE_fac_lm_10,"Vol-F10-LR")
PE_Vol_F20_LR <- cbind(PE_fac_lm_20,"Vol-F20-LR")
PE_Vol_F20_RR <- cbind(PE_fac_ridge_20,"Vol-F20-RR")



# plot Figure 2

library(ggplot2)
library(reshape2)
library(latex2exp)

box <- rbind(PE_Fun_LR, PE_Vol_RR, PE_Vol_F2_LR, PE_Vol_F10_LR, PE_Vol_F20_LR,PE_Vol_F20_RR)
box <- as.data.frame(box)
box[,1] <- as.numeric(as.character(box[,1]))
box[,2] <- as.factor(box[,2])
colnames(box) <- c("PE", "label")
box$label <- factor(box$label,levels =  c("Fun-LR", "Vol-RR", "Vol-F2-LR","Vol-F10-LR","Vol-F20-LR","Vol-F20-RR"))

pp <-ggplot(box,aes(x=label,y=PE,fill=label))+
  geom_boxplot(outlier.size=1,outlier.alpha=0.2)+ 
  xlab("")+ylab("PE")+ scale_color_manual(values=
                                            c("gray14","darkred","gray55","aquamarine",
                                              "chartreuse","royalblue","#56B4E9","#E69F00","tomato4"))+ylim(0.4,0.75) +
  theme(
    axis.title.x = element_text(size=20),
    axis.title.y = element_text(size=20),
    axis.text.y = element_text(size=20),
    axis.text.x = element_text(size=20),
    legend.title = element_text(size=20),
    legend.text = element_text(size=20),
    legend.position ="top"
  )
plot(pp)






# plot Figure 5
data <- cbind(MMSE,score,age1,as.numeric(gender1) )
q <- 3
K <- 2
Alpha_ad <- matrix(0,p,100)
A_ad <- array(0,c(p,T,100))
for (seed in 1:100) {
  set.seed(seed)
  index <- sample(1:n,0.95*n,replace = FALSE)
  data1 <- data[index,]
  data1 <- as.data.frame(data1)
  lm_result <- lm(MMSE~.,data1)
  beta_ad <- lm_result$coefficients[-1]
  
  eigenfun <- array(0,c(K*q,q,T))
  for (kk in 1 : K) {
    for (qq in 1 : q) {
      for (tt in 1:T) {
        eigenfun[ (qq-1)*K+kk,qq,tt] <- Theta_q[kk,,qq] %*% Bt[tt,]
      } 
    }
  }
  alpha_ad <- matrix(0,p,T)
  for (t in 1:T)  alpha_ad[,t] <- c((Bf) %*% solve(t(Bf) %*% (Bf)) %*% t(eigenfun[,,t]) %*% beta_ad[1:(q*K)]) 
  A_ad[,,seed] <- alpha_ad
  #print(seed)
  
}
A_ad_mean  <- matrix(0,p,T)
for (seed in 1 : 100) A_ad_mean <- A_ad_mean + A_ad[,,seed]/100


gg_data1 <- data.frame(x = time,y1 = A_ad_mean[7,],y2 = A_ad_mean[23,])
gg_data1[,-1] <-   apply(gg_data1[,-1],2,function(x) (ksmooth(gg_data1[,1],x,bandwidth = 0.1))$y )
a1 <- ggplot(gg_data1, aes(x = x))+geom_line(aes(y = y1, color = "left putamen"), size = 5) +geom_line(aes(y = y2, color = "right putamen"), size = 5)+
  scale_color_manual(values = c("left putamen" = "red","right putamen" = "blue"),labels=c("left putamen","right putamen"))+
  theme(axis.title.x = element_text(size=20),
        axis.title.y = element_text(size=20),
        axis.text.y = element_text(size=20),
        axis.text.x = element_text(size=20),
        legend.title = element_blank(),
        legend.text = element_text(size=20),
        legend.position = c(0.3,0.85))+xlab("quantile")+ylab(expression(italic(alpha(t))))+ guides(color = guide_legend(override.aes = list(size = 20)))
plot(a1)


gg_data1 <- data.frame(x = time,y1 = A_ad_mean[29,],y2 = A_ad_mean[30,])
gg_data1[,-1] <-   apply(gg_data1[,-1],2,function(x) (ksmooth(gg_data1[,1],x,bandwidth = 0.1))$y )
a2 <- ggplot(gg_data1, aes(x = x))+geom_line(aes(y = y1, color = "left basal forebrain"), size = 5) +geom_line(aes(y = y2, color = "right basal forebrain"), size = 5)+
  scale_color_manual(values = c("left basal forebrain" = "red","right basal forebrain" = "blue"),labels=c("left basal forebrain","right basal forebrain"))+
  theme(axis.title.x = element_text(size=20),
        axis.title.y = element_text(size=20),
        axis.text.y = element_text(size=20),
        axis.text.x = element_text(size=20),
        legend.title = element_blank(),
        legend.text = element_text(size=20),
        legend.position = c(0.3,0.85))+xlab("quantile")+ylab(expression(italic(alpha(t))))+ guides(color = guide_legend(override.aes = list(size = 20)))
plot(a2) 


gg_data1 <- data.frame(x = time,y1 = A_ad_mean[8,],y2 = A_ad_mean[24,])
gg_data1[,-1] <-   apply(gg_data1[,-1],2,function(x) (ksmooth(gg_data1[,1],x,bandwidth = 0.1))$y )
a3 <- ggplot(gg_data1, aes(x = x))+geom_line(aes(y = y1, color = "left pallidum"), size = 5) +geom_line(aes(y = y2, color = "right pallidum"), size = 5)+
  scale_color_manual(values = c("left pallidum" = "red","right pallidum" = "blue"),labels=c("left pallidum","right pallidum"))+
  theme(axis.title.x = element_text(size=20),
        axis.title.y = element_text(size=20),
        axis.text.y = element_text(size=20),
        axis.text.x = element_text(size=20),
        legend.title = element_blank(),
        legend.text = element_text(size=20),
        legend.position = c(0.3,0.85))+xlab("quantile")+ylab(expression(italic(alpha(t))))+ guides(color = guide_legend(override.aes = list(size = 20)))
plot(a3) 


gg_data1 <- data.frame(x = time,y1 = A_ad_mean[3,],y2 = A_ad_mean[19,])
gg_data1[,-1] <-   apply(gg_data1[,-1],2,function(x) (ksmooth(gg_data1[,1],x,bandwidth = 0.1))$y )
a4 <- ggplot(gg_data1, aes(x = x))+geom_line(aes(y = y1, color = "left cerebellum exterior"), size = 5) +geom_line(aes(y = y2, color = "right cerebellum exterior"), size = 5)+
  scale_color_manual(values = c("left cerebellum exterior" = "red","right cerebellum exterior" = "blue"),labels=c("left cerebellum exterior","right cerebellum exterior"))+
  theme(axis.title.x = element_text(size=20),
        axis.title.y = element_text(size=20),
        axis.text.y = element_text(size=20),
        axis.text.x = element_text(size=20),
        legend.title = element_blank(),
        legend.text = element_text(size=20),
        legend.position = c(0.3,0.85))+xlab("quantile")+ylab(expression(italic(alpha(t))))+ guides(color = guide_legend(override.aes = list(size = 20)))
plot(a4) 
